//---------------------------------------------------------------------------
//
// Evj        Model the evolution of DNA binding site recognition
//
// Module:    IconicLogoPanel, a Swing panel that displays a sequence logo.
// Author:    Paul C. Anagnostopoulos
// Created:   24 September 2005
//
// This computer program fits the description in the U.S. Copyright Act
// of a "United States Government work." It was written as a part of
// my/our official duties as a Government employee. This means it cannot
// be copyrighted. The program is freely available for use without a
// copyright notice, and there are no restrictions on its use, now or
// subsequently. We retain no rights in the program. Please acknowledge
// the source of this code in any releases you make.
//
//---------------------------------------------------------------------------


import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.lang.Math;
import java.util.Arrays;
import javax.swing.*;
import javax.swing.event.*;

//                      Gene Sequence Panel
//                      ---- -------- -----


public final class IconicLogoPanel extends JPanel implements NexusObserver
{

  private Nexus acceptedNexus = null;
  private boolean firstTime = true;

  public IconicLogoPanel ()
  {
    super();
  }

  // This public method implements the NexusObserver interface. It
  // simply saves the supplied nexus and repaints the display.

  public void acceptNexus (Nexus nexus)
  {
    acceptedNexus = nexus;
    repaint();
  }

//                      Constants and One-Time Values
//                      --------- --- -------- ------


  private static final int margin = 21,
                           tickLength = 4;

  private static final Font annoFont = new Font("SansSerif", Font.BOLD, 12),
                            baseFont = new Font("SansSerif", Font.BOLD, 20);

  private static int pageWidth, pageHeight;

  private static int logoWidth, logoHeight;

  private static int xOrigin = margin,
                     yOrigin;
  private static int xLogo = xOrigin + 1,
                     yLogo;

  private static int[] letterWidth = new int[4],
                       letterHeight = new int[4];

  private static int pixelsPerBit;

  private void firstTimeValues ()
  {
    pageWidth  = this.getWidth()  - margin;
    pageHeight = this.getHeight() - 2*margin;

    logoWidth  = pageWidth;
    logoHeight = pageHeight;

    yOrigin = pageHeight + margin;
    yLogo = yOrigin - 1;
    
    FontRenderContext frc = g2.getFontRenderContext();
    for (int b = 0; b <= 3; ++b) {
      GlyphVector gv = baseFont.createGlyphVector(frc, Base.letter(b));
      GlyphMetrics gm = gv.getGlyphMetrics(0);
      Rectangle2D bounds = gm.getBounds2D();
      letterWidth[b]  = (int)Math.round(bounds.getWidth());
      letterHeight[b] = (int)Math.round(bounds.getHeight());
    }

    pixelsPerBit = logoHeight / 2;
  }

//                      Paint Component
//                      ----- ---------


  private Graphics2D g2;

  private int[] comp = new int[4];
  private double[] freq = new double[4];
  private int[] order = new int[4];
  
  private int stackWidth;


  public void paintComponent (Graphics g)
  {
    super.paintComponent(g);

    // Pick up the nexus so we are impervious to nexuses coming in
    // while drawing. If it is null, forget it.

    Nexus nexus = acceptedNexus;
    if (nexus == null) return;
    g2 = (Graphics2D)g;
    
///    g2.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS,
///                        RenderingHints.VALUE_FRACTIONALMETRICS_OFF);
///    g2.setRenderingHint(RenderingHints.KEY_RENDERING,
///                        RenderingHints.VALUE_RENDER_SPEED);
///    g2.setRenderingHint(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING,
///                        java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);

    // Calculate values that we need determine only once.
    
    if (firstTime) {
      firstTimeValues();
      firstTime = false;
    }

    int[] site = nexus.sitePos;
    int sites = site.length;
    int siteWidth = nexus.siteWidth;
    byte[] chromosome = nexus.chromosome;
    
    stackWidth = (int)((float)logoWidth / siteWidth);
    
    drawAnnotation();
    
    // Iterate over the positions in the bindings sites.

    for (int p = 0; p < siteWidth; ++p) {

      // Count the occurences of each of the four bases.

      Arrays.fill(comp, 0);
      for (int s = 0; s < sites; ++s) ++comp[chromosome[site[s] + p]];

      // Calculate the frequency of each base and the uncertainty
      // of this position, Hs.

      double Hs = 0.0;
      for (int b = 0; b <= 3; ++b) {
        freq[b] = (double)comp[b] / sites;
        if (comp[b] > 0) Hs -= freq[b] * Stats.log2(freq[b]);
      }

      // Determine the order of the bases from least frequent to most.

      int b = 0;
      for (int n = 0; n <= 3; ++n) {
        for (int i = 0, least = Integer.MAX_VALUE; i <= 3; ++i) {
          if (comp[i] < least) {
            least = comp[i];
            b = i;
          }
        }
        order[n] = b;
        comp[b] = Integer.MAX_VALUE;
      }

      // Draw the letter stack for this position.

      drawStack(p, nexus.E_Hnb - Hs);
    }
  }

//                      Draw Annotation
//                      ---- ----------


  // This method draws the x- and y-axes and tick marks.
  
  private void drawAnnotation ()
  {
    g2.setFont(annoFont);

    g2.fillRect(xOrigin, yOrigin, pageWidth, 1);
    g2.fillRect(xOrigin, margin, 1, pageHeight);

    drawTick(margin,                                  "2.0");
    drawTick(margin + pixelsPerBit/2,                 "1.5");
    drawTick(margin + pixelsPerBit,                   "1.0");
    drawTick(margin + pixelsPerBit + pixelsPerBit/2,  "0.5");
    drawTick(yOrigin,                                 "0.0");
    
    g2.drawString("site positions", margin + pageWidth/3, yOrigin+margin-3);
    g2.drawString("bits", 0, margin-8);
  }

  private void drawTick (int y, String label)
  {
    g2.fillRect(xOrigin - tickLength, y, tickLength, 1);
    g2.drawString(label, 0, y+4);
  }

//                      Draw Stack
//                      ---- -----


  private void drawStack (int pos, double info)
  {
    int x = xLogo + pos * stackWidth;
    int y = yLogo;
    int yPixels;
    float xScale, yScale;

    g2.setFont(baseFont);

    for (int i = 0; i <= 3; ++i, y -= yPixels) {
      int b = order[i];

      xScale = (float)stackWidth / letterWidth[b];

      yPixels = (int)(freq[b] * info * pixelsPerBit);
      if (yPixels == 0) continue;
      yScale = (float)yPixels / letterHeight[b];
///Debug.trace("info", new Integer(pos), Base.letter(b), new Integer(y), new Integer(yPixels), new Float(yScale), new Double(info));

///      g2.translate(x, y);
///g2.setPaint(Color.black); g2.fillRect(0, 0, 2, 2); g2.fillRect(0, -yPixels, 2, 2);
      g2.scale(xScale, yScale);
      g2.setPaint(Base.color(b));
      g2.drawString(Base.letter(b), x/xScale, y/yScale);
      g2.scale(1.0f/xScale, 1.0f/yScale);
///      g2.translate(-x, -y);
    }
  }

}